
    .macro  mov_l  r_dest, r_src
#if defined (__AVR_HAVE_MOVW__)
    movw    \r_dest, \r_src
#else
    mov \r_dest, \r_src
#endif
    .endm

    .macro  mov_h  r_dest, r_src
#if defined (__AVR_HAVE_MOVW__)
    ; empty
#else
    mov \r_dest, \r_src
#endif
    .endm

#define __tmp_reg__ r0
#define __zero_reg__ r1


/*
*   Basic approach for clz:
*   Shift the value until a '1' is found.
*   Make sure the shifting stops by settings the carry bit
*   initially.
*
*   Since GCC does not know about the carry, there is no way GCC can
*   create such small routines
*
*   The focus is on the 16 bit counter since this is checked in the testsuite
*/
/*******************************************************
               Count leading zero 8 bit
*******************************************************/
#if defined (L_clzqi2)

#define r_arg1L r24     /* Low */
#define r_arg1H r25     /* High */
#define r_count __tmp_reg__ /* count result */

    .global __clzqi2
    .func   __clzqi2
__clzqi2:
    clr r_count     ; load with 0
    com r_count     ; invert (load with -1) + set carry
__clzqi2_loop:
    rol r_arg1L         ; Rotate through carry
    inc r_count         ; Carry not touch by inc
    brcc __clzqi2_loop   ; Branch on no carry
    mov r_arg1L, r_count
    clr r_arg1H         ; GCC expects int as return: we need the clear
    ret
#undef r_arg1L
#undef r_arg1H
#undef r_count
.endfunc

#endif

/*******************************************************
               Count leading zero 16 bit
*******************************************************/
#if defined (L_clzhi2)

#define r_arg1L r24     /* Low */
#define r_arg1H r25     /* High */
#define r_count __tmp_reg__

	.global	__clzhi2
	.func	__clzhi2
__clzhi2:
    clr r_count     ; load with 0
    com r_count     ; bit wise invert (set -1), set carry
__clzhi2_loop:
    rol r_arg1L
    rol r_arg1H
    inc r_count
    brcc __clzhi2_loop
    ; Setup return value
    mov r_arg1L, r_count
    clr r_arg1H                 ; GCC expects int as return: we need the clear
    ret

#undef r_arg1L
#undef r_arg1H
#undef r_count
.endfunc

#endif

/*
*   Implemented 32-bit clz as conditional 16 bit
*   If high word is non zero, then count only high word.
*   Else mov low word to high word and do 16 bit count
*   Add 16 for high word being all zero
*   This is faster for values < 64k.
*   It also save size if both clzsi and clzhi are use.
*
*   NOTE: if only clzsi is used, it is shorter to use an
*   make the same loop for 32-bit as for 16-bit. Using 4x a ROL/ADC
*
*/
/*******************************************************
               Count leading zero 32 bit
*******************************************************/
#if defined (L_clzsi2)

#define r_arg1L  r22
#define r_arg1H  r23
#define r_arg1HL r24
#define r_arg1HH r25

	.global	__clzsi2
	.func	__clzsi2
__clzsi2:
    adiw r_arg1HL, 0        ; test high word
    brne __clzhi2
                            ; High word == 0
    mov_l r_arg1HL, r_arg1L ; mov low word to high word
    mov_h r_arg1HH, r_arg1H
    rcall __clzhi2           ; Do a normal call, for L:H
    subi r_arg1HL, lo8(-16) ; add 16 for HL:HH being zero
                            ; r25 cleared by __clzhi2
    ret

#undef r_arg1L
#undef r_arg1H
#undef r_arg1HL
#undef r_arg1HH
#undef r_count
.endfunc

#endif







/*******************************************************
               Count trailing zero 16 bit
*******************************************************/
#if defined (L_ctzhi2)

#define r_arg1L r24     /* Low */
#define r_arg1H r25     /* High */
#define r_count __tmp_reg__

	.global	__ctzhi2
	.func	__ctzhi2
__ctzhi2:
    clr r_count ; load with 0
    com r_count     ; bit wise invert (set -1), set carry
__ctzhi2_loop:
    ror r_arg1H
    ror r_arg1L
    inc r_count
    brcc __ctzhi2_loop
    ; Setup return value
    mov r_arg1L, r_count
    clr r_arg1H                 ; GCC expects int as return: we need the clear
    ret

#undef r_arg1L
#undef r_arg1H
#undef r_count
.endfunc

#endif

/*******************************************************
               Count trailing zero 32 bit
*******************************************************/
#if defined (L_ctzsi2)

#define r_arg1L  r22
#define r_arg1H  r23
#define r_arg1HL r24
#define r_arg1HH r25
#define r_count __tmp_reg__

	.global	__ctzsi2
	.func	__ctzsi2
__ctzsi2:
    clr r_count ; load with 0
    com r_count ; bit wise invert (set -1), set carry
__ctzsi2_loop:
    ror r_arg1HH
    ror r_arg1HL
    ror r_arg1H
    ror r_arg1L
    inc r_count
    brcc __ctzsi2_loop
    ; Setup return value
    mov r_arg1HL, r_count
    clr r_arg1HH                 ; GCC expects int as return: we need the clear
    ret

#undef r_arg1L
#undef r_arg1H
#undef r_arg1HL
#undef r_arg1HH
#undef r_count

.endfunc

#endif






/*  POP COUNT
*
*   Global design goal: use as little as flash as possible!
*   Focus is on the 16-bit routine, since we got 16-bit int
*   We use a shift that counts the number of 1 that fell off during the shift
*
*/

/*******************************************************
               Count '1' in 8 bit
*******************************************************/
#if defined (L_popcountqi2)
#define r_loop   r20

	.global	__popcountqi2
	.func	__popcountqi2
__popcountqi2:
    ldi r_loop, 7
    rjmp __popcounthi2_clear_count
    .endfunc

#undef r_loop
#endif /* L_popcountqi2 */


/*******************************************************
               Count '1' in 16 bit
*******************************************************/
#if defined (L_popcounthi2)

#define r_loop   r20
#define r_count  __tmp_reg__
#define r_arg1L  r24
#define r_arg1H  r25

	.global	__popcounthi2
	.func	__popcounthi2
__popcounthi2:
    ldi r_loop, 16
__popcounthi2_clear_count:
    clr r_count ; load with 0
__popcounthi2_loop:
    rol r_arg1L
    rol r_arg1H
    adc r_count, __zero_reg__
    dec r_loop
    brne __popcounthi2_loop
    ; Setup return value
    mov r_arg1L, r_count
    clr r_arg1H                 ; GCC expects int as return: we need the clear
    ret

#undef r_loop
#undef r_count
#undef r_arg1L
#undef r_arg1H

.endfunc
#endif /* L_popcounthi2 */

/*******************************************************
               Count '1' in 32 bit
*******************************************************/
#if defined (L_popcountsi2)

#define r_loop   r20
#define r_count  __tmp_reg__
#define r_arg1L  r22
#define r_arg1H  r23
#define r_arg1HL r24
#define r_arg1HH r25

	.global	__popcountsi2
	.func	__popcountsi2
__popcountsi2:
    rcall __popcounthi2         ; do normal count on higher 16 bit
    ;It is shorter to setup for 16-bit loop count here
    ;It removed the need for the storing and adding of the result
    ;We also do not need to reload counter
    mov_l r_arg1HL, r_arg1L     ; move word
    mov_h r_arg1HH, r_arg1H
    ldi r_loop, 16              ; set loop length
    rjmp __popcounthi2_loop          ; jump to start of loop

#undef r_loop
#undef r_count
#undef r_arg1L
#undef r_arg1H
#undef r_arg1HL
#undef r_arg1HH

.endfunc
#endif /* L_popcountsi2 */

/*******************************************************
               Parity '1' in 16 bit
               return 0 = Even
               return 1 = Odd
*******************************************************/
#if defined (L_parityhi2)

#define r_arg1L r24
#define r_arg1H r25

	.global	__parityhi2
	.func	__parityhi2
__parityhi2:
    rcall __popcounthi2         ; do count
    andi r_arg1L, lo8(1)        ; only bit 0 is needed
                                ; r25 is already cleared by __popcounthi2
    ret

#undef r_arg1L
#undef r_arg1H

.endfunc
#endif /* L_parityhi2 */


/*******************************************************
               Parity '1' in 32 bit
               return 0 = Even
               return 1 = Odd
*******************************************************/
#if defined (L_paritysi2)

#define r_arg1L r24
#define r_arg1H r25

	.global	__paritysi2
	.func	__paritysi2
__paritysi2:
    rcall __popcountsi2         ; do count
    andi r_arg1L, lo8(1)        ; only bit 0 is needed
                                ; r25 is already cleared by __popcountsi2
    ret

#undef r_arg1L
#undef r_arg1H

.endfunc
#endif /* L_paritysi2 */

